import struct
import sys
import datetime
import argparse
import logging

def make_sector_list(num_sectors, skip: int, start_pos=0, first_sector=1):
    """ Create an array of sector positions """
    #
    # Have not found where InterMedia stores the interleave
    #
    if skip == 0: skip = 1
    if skip == 1 and start_pos == 0:
        return bytearray(range(first_sector, num_sectors + first_sector))
    ar = bytearray(num_sectors)
    for inx in range(num_sectors):
        ar[inx] = 255
    pos = start_pos
    for inx in range(num_sectors):
        while ar[pos] != 255:
            pos = (pos + 1) % num_sectors
        ar[pos] = inx + first_sector
        pos = (pos + skip) % num_sectors
    return ar


class Sector():

    def __init__(self, snum:int, data=None):
        self.sectorNum = snum
        self.data = data
        self.available = True
        self.compressed = False

class Track():

    head = 0
    cyl = 1
    mfm = 0  # Encoding
    rate = 0 # Transfer rate

    def __init__(self, cyl: int, head: int, mfm: int):
        self.cyl = cyl
        self.head = head
        self.mfm = mfm
        self.sector_size = 0
        self.sectors = []

    def append_sector(self, sector: Sector):
        """ Add a sector to the end of the track """
        self.sectors.append(sector)

    def insert_sector(self, pos: int, sector: Sector):
        self.sectors.insert(pos, sector)

    def remove_sector(self, sector: int):
        """ Remove the sector with physical index """
        del self.sectors[sector]

    def get_sector_list(self) -> bytearray:
        """ Create sector list from sector number in sectors """
        result = bytearray(len(self.sectors))
        i = 0
        for s in self.sectors:
            result[i] = s.sectorNum
            i += 1
        return result

    def get_num_sectors(self):
        """ Return number of physical sectors in track """
        return len(self.sectors)

class BINWriter():
    def __init__(self, fn, comment):
        self.outFile = open(fn, "wb")

    def setNumSectors(self, nsect):
        pass
    
    def set_sector_size(self, sector_size):
        self.sector_size = sector_size

    def setInterleave(self, nsect, il):
        """ Interleave is a number between 1 and X """
        pass

    def setSectorList(self, sectorList):
        pass

    def writeHead(self, comment):
        pass

    def writeCylHead(self, cyl, head, mfmFlag, transRate):
        pass

    def writeSector(self, sector):
        if not sector.available:
            self.outFile.write(bytes(self.sector_size))
        else:
            self.outFile.write(sector.data)

    def finishTrack(self):
        pass

    def close(self):
        self.outFile.close()


class IMDWriter():

    def __init__(self, fn, comment):
        self.outFile = open(fn, "wb")
        self.writeHead(comment)

    def setNumSectors(self, nsect):
        """ For current track """
        self.nsect = nsect

    sector_sizeTable = { 128:0, 256:1, 512:2, 1024:3, 2048:4, 4096:5, 8192:6 }

    def set_sector_size(self, sector_size):
        self.size_code = self.sector_sizeTable[sector_size]

    def setSizeCode(self, size_code):
        self.size_code = size_code

    def setInterleave(self, nsect, il):
        """ Interleave is a number between 1 and X """
        self.il = il
        self.sectorList = make_sector_list(nsect, il)

    def setSectorList(self, sectorList):
        self.sectorList = sectorList

    def writeHead(self, comment):
        """ Write IMD header and ASCII comment terminated by Ctrl+Z """
        if not comment.startswith('IMD ') and not comment.startswith('\n'):
            comment = '\n' + comment
        cmtLines = comment.split('\n')
        dateStr = datetime.datetime.now().strftime("%d/%m/%Y %H:%M:%S")
        cmtLines[0] = "IMD 1.18: %s" % dateStr
        bytecmt = '\r\n'.join(cmtLines).encode('ascii')
        self.outFile.write(bytecmt)
        if len(bytecmt) != 0 and not bytecmt.endswith(b'\r\n'):
            self.outFile.write(b'\r\n')
        self.outFile.write(b'\x1A')

    def writeCylHead(self, cyl, head, mfmFlag, transRate):
        modeCode = self._getModeCode(mfmFlag, transRate)
        self.outFile.write(bytes((modeCode, cyl, head, self.nsect, self.size_code)))
        self.outFile.write(self.sectorList)
        self.sectorDict = {}

    modeTable = { (0, 500):0, (0,300):1, (0,250):2, (1,500):3, (1,300):4, (1,250):5 } 

    def _getModeCode(self, mfmFlag, transRate):
        return self.modeTable[(mfmFlag, transRate)]
        #if mfmFlag == 1: return 4
        #else: return 0  # 300 kpbs FM

    def _isCompressible(self, data):
        first = data[0]
        for n in data:
            if n != first: return False
        return True

    def writeSector(self, sector):
        self.sectorDict[sector.sectorNum] = sector

    def finishTrack(self):
        for s in self.sectorList:
            sector = self.sectorDict[s]
            self._wSector(sector)

    def _wSector(self, sector):
        if not sector.available:
            self.outFile.write(b'\x00')
            return
        if self._isCompressible(sector.data):
            self.outFile.write(b'\x02')
            self.outFile.write(bytes([sector.data[0]]))
        else:
            self.outFile.write(b'\x01')
            self.outFile.write(sector.data)

    def close(self):
        self.outFile.close()

class FloppyImage():
    def __init__(self):
        self.ncyls = 0
        self.nheads = 0
        self.tracks = []
        self.comment = ''
    
    def getSector(self, head, cyl, sector: int):
        for t in self.tracks:
            if t.head == head and t.cyl == cyl:
                for s in range(t.get_num_sectors()):
                    if t.sectors[s].sectorNum == sector:
                        return t.sectors[s]
        raise ValueError('Sector not found %d/%d/%d' % (head, cyl, sector))

    def get_max_cyl_num(self):
        """ Returns the highest cylinder number.
            Tracks start at number 1 and there are usually two tracks per
            cylinder on a double-sided disk.
        """
        return self.ncyls
        
    def get_track(self, head, cyl):
        for t in self.tracks:
            if t.head == head and t.cyl == cyl:
                return t
        raise ValueError('Sector not found %d/%d' % head, cyl)

    def get_capacity(self) -> int:
        """ Return the number of bytes that the image can contain. """
        diskSize = 0
        for t in self.tracks:
            diskSize += t.get_num_sectors() * t.sector_size
        return diskSize

    def changeData(self, head, cyl, sector, newdata):
        sectorRec = self.getSector(head, cyl, sector)
        sectorRec.data = newdata
        sectorRec.good = True
        sectorRec.available = True

    def append_track(self, track):
        self.tracks.append(track)
        self.ncyls = max(self.ncyls, track.cyl + 1)
        self.nheads = max(self.nheads, track.head + 1)

    def save(self, writer):
        for t in self.tracks:
            writer.set_sector_size(t.sector_size)
            writer.setNumSectors(t.get_num_sectors())
            sectorList = t.get_sector_list()
            writer.setSectorList(sectorList)
            writer.writeCylHead(t.cyl, t.head, t.mfm, t.rate)
            for sNum in sorted(sectorList):
                sector = self.getSector(t.head, t.cyl, sNum)
                writer.writeSector(sector)
            writer.finishTrack()
        writer.close()

def imdload(disk: FloppyImage, fn: str):
    """ Loads IMD disk """
    with open(fn, "rb") as fd:
        img = fd.read()
    if not img.startswith(b'IMD '):
        raise ValueError('Not an IMD disk')
    track_start = img.index(0x1A)
    disk.comment = img[:track_start].decode()
    disk.comment = disk.comment.replace('\x0D\x0A', '\n')
    track_start += 1 # Skip EOF char
    while track_start < len(img):
        modeValue, cyl, head, nsectors, sectCode = \
            img[track_start:track_start+5]
        track_start += 5
        mfm = (0, 0, 0, 1, 1, 1)[modeValue]
        t = Track(cyl, head, mfm)
        #print("C:{},H:{}".format(cyl, head))
        t.rate = (500, 300, 250, 500, 300, 250)[modeValue]
        t.sector_size = (128, 256, 512, 1024, 2048, 4096, 8192)[sectCode]
        sectorList = img[track_start:track_start + nsectors]
        track_start += nsectors
        for s in range(nsectors):
            sector = Sector(sectorList[s])
            dataRec = img[track_start]
            #print(s, " ", dataRec, end="_")
            sector.available = True
            track_start += 1
            sector.good = (False, True, True, False, False, False, False, False, False)[dataRec]
            if dataRec in (2, 4, 6, 8):
                sector.data = bytearray(t.sector_size)
                for i in range(t.sector_size):
                    sector.data[i] = img[track_start]
                    sector.compressed = True
                track_start += 1
            elif dataRec in (1, 3, 5, 7):
                sector.data = img[track_start:track_start + t.sector_size]
                sector.compressed = False
                track_start += t.sector_size
            else:
                sector.available = False
                sector.data = None
            t.append_sector(sector)
        disk.append_track(t)

